local version = 1

if version < towerofheaven.version then
	return -- Don't load the rest of the script.
end

-- Cache all of the font glyphs before anything...
local tohFont = nil
local newLineChar = string.byte('\n')
local newLineHeight = 9

local bgBackgroundPalette = 178
local bgBorder1Palette = 108
local bgBorder2Palette = 110

towerofheaven.animTime = TICRATE/2

local hAlignments = {"left", "center", "right"}
local yAlignments = {"top", "center", "bottom"}

local function cacheFont(v)
	tohFont = {}
	
	local from, to = 32, 126
	local fontname = "TOHF"
	
	for i = from, to do	
		local fontGraphic = string.format("%s%03d", fontname, i)
		table.insert(tohFont, i, v.cachePatch(fontGraphic))
	end
end

local function powerOf(num, to)
	local final = num
	for i = 1, to do final = FixedMul(final, num) end
	return final
end

local function easeOut(a)
	local frac = FRACUNIT - a
	return FRACUNIT - powerOf(frac, 2)
end

local function snapToNearestPixel(num)
	return (num>>FRACBITS)<<FRACBITS
end

local function turnNewLineSeparatedStringIntoTable(string)
	if string == nil or string == "" then
		return {}, 0
	end

	local tableVar = {}
	local newString = ""
	local numThingsAdded = 0
	for i = 1, #string do
		local c = string:sub(i,i)
		
		if c == '\n' then			
			table.insert(tableVar, newString)
			numThingsAdded = $ + 1
			newString = ""
		else
			newString = $..c
		end
	end

	if newString ~= "" then
		table.insert(tableVar, newString)
		numThingsAdded = $ + 1
	end

	return tableVar, numThingsAdded
end

local function getTextWidthAndHeight(strings, numStrings)
	local biggestWidth = 0
	
	for i = 1, numStrings do
		local iterations = string.len(strings[i])
		if iterations <= 0 then continue end
		local text = strings[i]
		local currentWidth = 0
		
		for j = 1, iterations do
			local chara = string.sub(text, j, j)
			if chara == '\n' then continue end
			local asciiCode = string.byte(chara)
			local sprite = tohFont[asciiCode]
		
			currentWidth = $ + sprite.width
		end
		
		if currentWidth > biggestWidth then biggestWidth = currentWidth end
	end
	
	return biggestWidth<<FRACBITS, (newLineHeight * numStrings)<<FRACBITS
end

local function getStringWidthAndHeight(string)
	return getTextWidthAndHeight({string}, 1)
end

local function sortWidthAlignment(obj, stringNum)
	local currentWidth = 0

	if obj.alignH == "center" then
		local thisString = obj.messageTable[stringNum]
		local thisLineWidth = getStringWidthAndHeight(thisString)
		currentWidth = FixedDiv(obj:getWidth() - FixedMul(thisLineWidth, obj.scale), 2<<FRACBITS)
	--[[
	elseif obj.alignH == "right" then 
		-- TODO: probably would need more than this function to work.
	]]
	end
	
	return currentWidth
end

towerofheaven.initToHString = function(v, message, posX, posY, alignHorizontal, alignVertical, flags, scale)
	if not tohFont then cacheFont(v) end
	
	local messageObj = {}
	
	messageObj.fullMessage = message
	messageObj.x, messageObj.y = posX, posY
	messageObj.flags = flags
	messageObj.scale = scale
	messageObj.alignH, messageObj.alignV = alignHorizontal, alignVertical
	messageObj.messageTable, messageObj.messageNum = turnNewLineSeparatedStringIntoTable(messageObj.fullMessage)
	messageObj.prevWidth, messageObj.prevHeight = 0, 0
	messageObj.lastChangedMsgTic = leveltime
	messageObj.width, messageObj.height = getTextWidthAndHeight(messageObj.messageTable, messageObj.messageNum)
	messageObj.displayText = function(self)
		if not (self.fullMessage and string.len(self.fullMessage) > 0) then return end
		local x, y = self.x<<FRACBITS, self.y<<FRACBITS
	
		if self.alignH == "center" then
			x = $ - FixedDiv(self:getWidth(), 2<<FRACBITS)
		elseif self.alignH == "right" then
			x = $ - self:getWidth()
		end
		
		if self.alignV == "center" then
			y = $ - FixedDiv(self:getHeight(), 2<<FRACBITS)
		elseif self.alignV == "bottom" then
			y = $ - self:getHeight()
		end
	
		local iterations = string.len(self.fullMessage)
		local currentWidth = 0
		local currentHeight = 0
		local currentLine = 1
		
		currentWidth = sortWidthAlignment(self, currentLine)
		
		for i = 1, iterations do
			local chara = string.sub(self.fullMessage, i, i)
			local asciiCode = string.byte(chara)
			
			if asciiCode == newLineChar then -- Newline
				currentLine = $ + 1
				currentHeight = $ + FixedMul(newLineHeight<<FRACBITS, self.scale)
				currentWidth = sortWidthAlignment(self, currentLine)

				continue
			end
			
			local sprite = tohFont[asciiCode]
			v.drawScaled(snapToNearestPixel(x + currentWidth), snapToNearestPixel(y + currentHeight), self.scale, sprite, self.flags)
			currentWidth = $ + FixedMul(sprite.width<<FRACBITS, self.scale)
		end
	end
	messageObj.displayBG = function(self)
		-- Always make it two pixels wider and height-ier
		local x = (self.x - 2)<<FRACBITS
		local width = self:getWidth() + (4<<FRACBITS)
		
		if self.alignH == "center" then
			x = $ - FixedDiv(self:getWidth(), 2<<FRACBITS)
		elseif self.alignH == "right" then
			x = $ - self:getWidth()
		end
		
		local y = (self.y - 2)<<FRACBITS
		local height = self:getHeight() + (3<<FRACBITS)
		
		if self.alignV == "center" then
			y = $ - FixedDiv(self:getHeight(), 2<<FRACBITS)
		elseif self.alignV == "bottom" then
			y = $ - self:getHeight()
		end
		
		x, y, height, width = x>>FRACBITS, y>>FRACBITS, height>>FRACBITS, width>>FRACBITS
		
		v.drawFill(x-2, y+height+2, width+4, 1, bgBorder1Palette|self.flags)
		v.drawFill(x-2, y-2, width+4, height+4, bgBorder2Palette|self.flags)
		v.drawFill(x-1, y-1, width+2, height+2, bgBorder1Palette|self.flags)
		v.drawFill(x, y, width, height, bgBackgroundPalette|self.flags)
	end
	messageObj.display = function(self)
		self:displayBG()
		self:displayText()
	end
	messageObj.getWidth = function(self)
		if leveltime >= self.lastChangedMsgTic + towerofheaven.animTime then
			return FixedMul(messageObj.width, self.scale)
		end
		
		local animMult = easeOut((FRACUNIT/towerofheaven.animTime)*(leveltime-messageObj.lastChangedMsgTic))
		local wDiff = self.width - self.prevWidth
		
		return FixedMul(self.prevWidth + FixedMul(wDiff, animMult), self.scale)
	end
	messageObj.getHeight = function(self)
		if leveltime >= self.lastChangedMsgTic + towerofheaven.animTime then
			return FixedMul(messageObj.height, self.scale)
		end
		
		local animMult = easeOut((FRACUNIT/towerofheaven.animTime)*(leveltime-messageObj.lastChangedMsgTic))
		local hDiff = self.height - self.prevHeight
		
		return FixedMul(self.prevHeight + FixedMul(hDiff, animMult), self.scale)
	end
	messageObj.changeText = function(self, text)
		self.fullMessage = text
		self.messageTable, self.messageNum = turnNewLineSeparatedStringIntoTable(self.fullMessage)
		self.prevWidth, self.prevHeight = self.width, self.height
		self.width, self.height = getTextWidthAndHeight(self.messageTable, self.messageNum)	
		
		self.lastChangedMsgTic = leveltime
	end
	messageObj.performDisplayLogic = function(self)
		if leveltime < self.lastChangedMsgTic + (towerofheaven.animTime/4)*3 then
			self:displayBG()
		elseif self.fullMessage ~= "" then
			self:display()
		end
	end

	return messageObj
end 