local version = 1

if version < towerofheaven.version then
	return -- Don't load the rest of the script.
end

towerofheaven.towerOfHeavenActive = false
local currentRules = {}

local preThinkFrame = "PreThinkFrame"
local thinkFrame = "ThinkFrame"
local mobjMoveCollide = "MobjMoveCollide"
local mobjCollide = "MobjCollide"

local starttime = 6*TICRATE + (3*TICRATE/4)

towerofheaven.playedThisMapHowManyTimes = 0
towerofheaven.thisMap = 0
towerofheaven.previousMap = 0

towerofheaven.isRuleActive = function(rule, player)
	local where = INT32_MAX
	for i = 1, #currentRules do
		if currentRules[i] and currentRules[i].name == rule then
			where = i
			break
		end
	end
	
	if player.laps >= where-1 and player.rulesLeniencyTimer == 0 then
		return true
	end

	return false
end

local function runRulesCode(player, hookType, extra1)
	local currentLap = player.laps
	
	for i = 1, currentLap do
		if not currentRules[i] then continue end -- Mostly for skipping the first lap's "rule" which is none.
		if player.rulesLeniencyTimer > 0 and i == currentLap then continue end -- Give the player some time to read the rule.
		local ruleName = currentRules[i].name
		if not towerofheaven.rulesLogic[ruleName][hookType] then continue end -- Nothing for this rule to execute here.
	
		-- Run the hook, let it rip.
		local parameters = currentRules[i].params
		towerofheaven.rulesLogic[ruleName][hookType](player, parameters, extra1)
	end
end

local function towerOfHeavenThinker(player)
	if leveltime == starttime then
		towerofheaven.performStartMessage(player)
	end

	player.previousLap = $ or 0
	player.rulesLeniencyTimer = $ or 0
	
	if player.rulesLeniencyTimer > 0 then
		player.rulesLeniencyTimer = $ - 1
	end
	
	if player.laps > player.previousLap and currentRules[player.laps] then -- New lap.
		towerofheaven.performGodMessage(player, player.laps, towerofheaven.getRuleDescription(currentRules[player.laps]))
		player.rulesLeniencyTimer = towerofheaven.stayForLong + towerofheaven.animTime
	end
	
	player.previousLap = player.laps
end

towerofheaven.isDamaged = function(player)
	local pks = player.kartstuff

	return player.powers[pw_flashing] > 0 or pks[k_squishedtimer] > 0 or pks[k_spinouttimer] > 0 or
			pks[k_invincibilitytimer] > 0 or (pks[k_growshrinktimer] > 0 or (player.hugequest and player.hugequest.huge > 0)) or pks[k_hyudorotimer] > 0 or
			(gametype == GT_MATCH and ((pks[k_bumper] <= 0 and pks[k_comebacktimer]) or pks[k_comebackmode] == 1))

end

local function thinkFrameThinker(typeOfThinker)
	if not towerofheaven.towerOfHeavenActive then return end
	
	for player in players.iterate do
		if not (player and player.valid and player.mo and player.mo.valid and not player.spectating) then continue end
		if typeOfThinker == thinkFrame then towerOfHeavenThinker(player) end
		
		if towerofheaven.isDamaged(player) then continue end
		runRulesCode(player, typeOfThinker)
	end
end

local function collisionThinker(typeOfThinker, tmthing, thing)
	if not towerofheaven.towerOfHeavenActive then return nil end
	
	-- Lua hot potato
	if not (thing and thing.valid) then return nil end
	if not (tmthing and tmthing.valid) then return nil end
	
	-- Height checks.
	if tmthing.momz < 0 then
		if tmthing.z + tmthing.momz > thing.z + thing.height then
			return false
		end
	elseif tmthing.z > thing.z + thing.height then
		return false
	end
	
	if tmthing.momz > 0 then
		if tmthing.z + tmthing.height + tmthing.momz < thing.z then
			return false
		end
	elseif tmthing.z + tmthing.height < thing.z then
		return false
	end
	
	local player = tmthing.player
	if not (player and player.valid) then return nil end
	if towerofheaven.isDamaged(player) then return nil end
	
	runRulesCode(player, typeOfThinker, thing)
end

addHook(thinkFrame, do 
	if version < towerofheaven.version then return end
	thinkFrameThinker(thinkFrame)
end)
addHook(preThinkFrame, do
	if version < towerofheaven.version then return end
	thinkFrameThinker(preThinkFrame)
end)
addHook(mobjMoveCollide, function(tmthing, thing)
	if version < towerofheaven.version then return end
	collisionThinker(mobjMoveCollide, tmthing, thing)
end, MT_PLAYER)
addHook(mobjCollide, function(thing, tmthing)
	if version < towerofheaven.version then return end
	collisionThinker(mobjCollide, thing, tmthing)
end, MT_PLAYER)

local function runParametersLogic(rule, fixedChoice)
	fixedChoice = $ or nil

	local ruleParameters = rule.params
	if ruleParameters then
		if ruleParameters.choices then
			local key = -1
			if fixedChoice and fixedChoice > 0 then
				key = fixedChoice
			else
				key = P_RandomRange(1, #ruleParameters.choices)
			end
			
			ruleParameters.currentChoice = ruleParameters.choices[key]
			if ruleParameters.strings then
				rule.descriptionChosen = string.format(rule.description, ruleParameters.strings[key])
			end
		end
	end
end

local function selectRandomRule(ruleTable)
	while #ruleTable > 0 do
		local key = P_RandomRange(1, #ruleTable)
		local chosenRule = ruleTable[key]
		table.remove(ruleTable, key)
		local validFunc = towerofheaven.rulesValidIf[chosenRule.name]
		if validFunc and not validFunc() then continue end
		runParametersLogic(chosenRule)
		return chosenRule
	end
end

local function selectRandomRuleWeighted(fraction, ruleTableOne, ruleTableTwo)
	if fraction > P_RandomFixed() then
		return selectRandomRule(ruleTableTwo)
	else
		return selectRandomRule(ruleTableOne)
	end
end

local cv_numlaps = nil
local cv_basenumlaps = nil
local function getThoseVars()
	if not cv_numlaps then
		cv_numlaps = CV_FindVar("numlaps")
	end
	
	if not cv_basenumlaps then
		cv_basenumlaps = CV_FindVar("basenumlaps")
	end
end

addHook("MapLoad", function(mapnum)
	if version < towerofheaven.version then return end
	if not mapheaderinfo[mapnum]["towerofheaven"] then
		towerofheaven.towerOfHeavenActive = false
		return
	end
	towerofheaven.towerOfHeavenActive = true
	getThoseVars()
	
	currentRules = {nil}
	
	local fixedRules = mapheaderinfo[mapnum]["tohfixedrarules"]
	if (modeattacking or towerofheaven.pInGame() <= 1) and fixedRules then
		local fixedRulesTable = towerofheaven.turnCommaDelimitedStringIntoTable(fixedRules)
		local fixedRulesChoicesTable = towerofheaven.turnCommaDelimitedStringIntoTable(mapheaderinfo[mapnum]["tohfixedraruleschoice"])
		for i = 1, #fixedRulesTable do
			local rule = towerofheaven.getRuleByName(fixedRulesTable[i])
			table.insert(currentRules, rule)
			runParametersLogic(rule, tonumber(fixedRulesChoicesTable[i]))
		end
	else -- Normal logic.
		local numLaps = max(max(mapheaderinfo[mapnum].numlaps, cv_numlaps.value), cv_basenumlaps.value)
		local numRules = numLaps - 1
		
		local rulesDifficultyWeight = FRACUNIT
		local rulesDifficultyLapFraction = (2<<FRACBITS)/(numRules-1)
		local percentageFraction = FRACUNIT/255
		
		local easyRules = towerofheaven.getRulesTable(towerofheaven.easyDifficulty)
		local normalRules = towerofheaven.getRulesTable(towerofheaven.normalDifficulty)
		local hardRules = towerofheaven.getRulesTable(towerofheaven.hardDifficulty)
		
		-- Ok, let's choose the rules.
		for i = 1, numRules do
			local rulesDiffInt = FixedInt(rulesDifficultyWeight)
			local chosenRule = nil
		
			if rulesDiffInt >= towerofheaven.hardDifficulty then -- Hard difficulty.
				chosenRule = selectRandomRule(hardRules)
			elseif rulesDiffInt >= towerofheaven.normalDifficulty then -- Normal difficulty.
				if rulesDifficultyWeight == towerofheaven.normalDifficulty<<FRACBITS then
					chosenRule = selectRandomRule(normalRules)
				else -- Pick something between normal and hard if the number isn't exact.
					local fraction = rulesDifficultyWeight - (towerofheaven.normalDifficulty<<FRACBITS)
					chosenRule = selectRandomRuleWeighted(fraction, normalRules, hardRules)
				end
			elseif rulesDiffInt >= towerofheaven.easyDifficulty then -- Easy difficulty.
				if rulesDifficultyWeight == towerofheaven.easyDifficulty<<FRACBITS then
					chosenRule = selectRandomRule(easyRules)
				else -- Pick something between easy and normal if the number isn't exact.
					local fraction = rulesDifficultyWeight - (towerofheaven.easyDifficulty<<FRACBITS)
					chosenRule = selectRandomRuleWeighted(fraction, easyRules, normalRules)
				end
			end
			
			table.insert(currentRules, chosenRule)
			
			rulesDifficultyWeight = $ + rulesDifficultyLapFraction
		end
	end
	
	towerofheaven.loadRuleInits()
	towerofheaven.setUpHitFeed()
	towerofheaven.playedThisMapHowManyTimes = $ + 1
end)

addHook("MapChange", function(mapnum)
	if version < towerofheaven.version then return end

	towerofheaven.previousMap = towerofheaven.thisMap
	towerofheaven.thisMap = mapnum
end)

addHook("NetVars", function(network)
	if version < towerofheaven.version then return end

	towerofheaven.towerOfHeavenActive = network($)
	towerofheaven.playedThisMapHowManyTimes = network($)
	towerofheaven.previousMap = network($)
	
	if not towerofheaven.towerOfHeavenActive then return end

	-- Yeah, it's not a good idea to pass *functions* through netvars huh, what a dummy I am.
	currentRules = network($)
end)