local GRUB_VERSION = 1

-- avoid redefiniton on updates
if grubmobj == nil then rawset(_G, "grubmobj", {}) end
if grubmobj.GRUB_VERSION == nil or grubmobj.GRUB_VERSION < GRUB_VERSION then

freeslot("SPR_GRGR", "SPR_FRGR", "SPR_GRDR",
	"MT_DIGGLE_SKD_GRUB",
	"sfx_grubd",
	"S_DIGGLE_SKD_GRUBMOVE", "S_DIGGLE_SKD_GRUBSLIDE", "S_DIGGLE_SKD_GRUBWEE")

mobjinfo[MT_DIGGLE_SKD_GRUB] = {
	--$Name Shovel Knight Dig Grub
	--$Sprite GRGRC0
	--$Category Diggle
	doomednum 	= 3901,
	spawnstate 	= S_DIGGLE_SKD_GRUBMOVE,
	meleestate 	= S_DIGGLE_SKD_GRUBSLIDE,
	seestate 	= S_DIGGLE_SKD_GRUBWEE,
	activesound = sfx_grubd,
	speed 		= 8<<FRACBITS,
	radius 		= 30<<FRACBITS,
	height 		= 16<<FRACBITS,
	flags 		= MF_SPECIAL|MF_BOUNCE,
}

local function determineSprite(actor)
	if actor.cusval - actor.extravalue2 > 0 then actor.sprite = SPR_GRGR
	else actor.sprite = SPR_FRGR end
end

function A_Grub_DetermineSpriteAndAngle(actor, var1, var2)
	actor.extravalue1 = actor.angle
	
	local finalAngle = AngleFixed(actor.cusval)
	
	if P_RandomChance(FRACUNIT/2) then finalAngle = $ + (FRACUNIT*90)
	else finalAngle = $ + (-FRACUNIT*90) end
	
	finalAngle = $ + ((FRACUNIT*45) - (P_RandomKey(91) * FRACUNIT))
	actor.extravalue2 = FixedAngle(finalAngle)
	
	
	determineSprite(actor)
end

function A_Grub_DetermineSprite(actor, var1, var2)
	determineSprite(actor)
end

states[S_DIGGLE_SKD_GRUBMOVE] = {
	sprite = SPR_GRGR,
	frame = FF_ANIMATE|A,
	tics = 15,
	var1 = 4,
	var2 = 3,
	action = A_Grub_DetermineSpriteAndAngle,
	nextstate = S_DIGGLE_SKD_GRUBSLIDE
}

states[S_DIGGLE_SKD_GRUBSLIDE] = {
	sprite = SPR_GRGR,
	frame = F,
	tics = 15,
	action = A_Grub_DetermineSprite,
	nextstate = S_DIGGLE_SKD_GRUBMOVE
}

states[S_DIGGLE_SKD_GRUBWEE] = {
	sprite = SPR_GRDR,
	frame = A,
	tics = -1,
	nextstate = S_DIGGLE_SKD_GRUBWEE
}

addHook("MobjThinker", function(mo)
	if mo.extravalue1 == 0 then
		mo.cusval = mo.angle
		mo.extravalue1 = 1
	end

	if mo.info.seestate ~= nil and mo.state == mo.info.seestate then
	
		local speed = R_PointToDist2(0, 0, mo.momx, mo.momy)
		local frame = (mo.frame & FF_FRAMEMASK)
		if speed > FixedMul(10*mo.scale, mapobjectscale) then
			frame = $ + 2
		elseif speed > FixedMul(5*mo.scale, mapobjectscale) and leveltime % 2 == 0 then
			frame = $ + 2
		elseif speed > FixedMul(2*mo.scale, mapobjectscale) and leveltime % 2 == 0 then
			frame = $ + 1
		end
		if frame > D then frame = $ % D end
		mo.frame = ($ & ~FF_FRAMEMASK)|frame
		
		if (abs(mo.momx) < mo.scale and abs(mo.momy) < mo.scale) and
		P_IsObjectOnGround(mo) then
			mo.cusval = R_PointToAngle2(0, 0, mo.momx, mo.momy)
			mo.friction = (FRACUNIT/100)*95
			mo.momx = 0
			mo.momy = 0
			mo.state = mo.info.spawnstate
		end

	elseif mo.info.meleestate ~= nil and mo.state == mo.info.meleestate
		and mo.tics == states[mo.info.meleestate].tics then
		
		mo.friction = (FRACUNIT/4)*1
		P_Thrust(mo, mo.extravalue2, FixedMul(mo.info.speed, mapobjectscale))
	end
	
end, MT_DIGGLE_SKD_GRUB)

addHook("TouchSpecial", function(special, toucher)
	if special.type == toucher.type then return true end

	local flip = P_MobjFlip(special)
	if (toucher.momx or toucher.momy) and flip * special.momz <= 0 then
	
		special.momx = toucher.momx
		special.momy = toucher.momy
		P_SetObjectMomZ(special, max(FixedHypot(toucher.momx, toucher.momy) / 4, FixedMul(special.info.speed, special.scale)), true)
		
		if flip * toucher.momz > 0 then
			special.momz = $ + FixedMul(toucher.momz / 8, mapobjectscale)
		end
		
		if special.info.seestate ~= nil then
			special.state = special.info.seestate
		end
		
		S_StartSound(special, special.info.activesound)
	end
	
	return true
end, MT_DIGGLE_SKD_GRUB)

addHook("MobjSpawn", function(mobj)
	P_SpawnShadowMobj(mobj)
end, MT_DIGGLE_SKD_GRUB)

grubmobj.GRUB_VERSION = GRUB_VERSION

end